from __future__ import annotations

from typing import TYPE_CHECKING, Callable, List

from sila2.features.silaservice import SiLAServiceBase, UnimplementedFeature
from sila2.framework.errors.no_metadata_allowed import NoMetadataAllowed

if TYPE_CHECKING:
    from sila2.server.sila_server import SilaServer


def no_metadata_allowed(func: Callable):
    def wrapper(*args, **kwargs):
        metadata = kwargs.pop("metadata")
        if metadata:
            raise NoMetadataAllowed(f"SiLAService.{func.__name__} received metadata {list(metadata.keys())}")
        return func(*args, **kwargs)

    return wrapper


class SiLAServiceImpl(SiLAServiceBase):
    parent_server: SilaServer

    def __init__(self, parent_server: SilaServer):
        self.parent_server = parent_server

    @no_metadata_allowed
    def GetFeatureDefinition(self, FeatureIdentifier: str) -> str:
        if FeatureIdentifier not in (f.fully_qualified_identifier for f in self.parent_server.features.values()):
            raise UnimplementedFeature(f"Feature {FeatureIdentifier} is not implemented by this server")

        return self.parent_server.features[FeatureIdentifier.split("/")[-2]]._feature_definition

    @no_metadata_allowed
    def SetServerName(self, ServerName: str) -> None:
        self.parent_server.server_name = ServerName

    @no_metadata_allowed
    def get_ImplementedFeatures(self) -> List[str]:
        return [f.fully_qualified_identifier for f in self.parent_server.features.values()]

    @no_metadata_allowed
    def get_ServerName(self) -> str:
        return self.parent_server.server_name

    @no_metadata_allowed
    def get_ServerType(self) -> str:
        return self.parent_server.server_type

    @no_metadata_allowed
    def get_ServerUUID(self) -> str:
        return str(self.parent_server.server_uuid)

    @no_metadata_allowed
    def get_ServerDescription(self) -> str:
        return self.parent_server.server_description

    @no_metadata_allowed
    def get_ServerVersion(self) -> str:
        return self.parent_server.server_version

    @no_metadata_allowed
    def get_ServerVendorURL(self) -> str:
        return self.parent_server.server_vendor_url



import os
from os.path import dirname, join

# https://github.com/protocolbuffers/protobuf/issues/3002#issuecomment-325459597
os.environ["PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION"] = "python"

resource_dir = join(dirname(__file__), "resources")



from __future__ import annotations

import logging
import re
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Optional, Tuple, Union

import grpc
from grpc import ServicerContext

from sila2.framework.abc.sila_error import SilaError
from sila2.framework.command.command import Command
from sila2.framework.defined_execution_error_node import DefinedExecutionErrorNode
from sila2.framework.errors.defined_execution_error import DefinedExecutionError
from sila2.framework.errors.invalid_metadata import InvalidMetadata
from sila2.framework.errors.undefined_execution_error import UndefinedExecutionError
from sila2.framework.errors.validation_error import ValidationError
from sila2.framework.fully_qualified_identifier import FullyQualifiedIdentifier
from sila2.framework.metadata import Metadata
from sila2.framework.property.property import Property
from sila2.framework.utils import FullyQualifiedIdentifierRegex, raise_as_rpc_error

if TYPE_CHECKING:
    from sila2.server.sila_server import SilaServer

logger = logging.getLogger(__name__)


def _find_allowed_defined_execution_errors(
    server: SilaServer,
    origin: Optional[Union[Command, Property]] = None,
    metadata_identifiers: Optional[Iterable[FullyQualifiedIdentifier]] = None,
) -> List[DefinedExecutionErrorNode]:
    allowed_errors = []
    if origin is not None:
        allowed_errors.extend(origin.defined_execution_errors)
    if metadata_identifiers is not None:
        for metadata_id in metadata_identifiers:
            allowed_errors.extend(server.children_by_fully_qualified_identifier[metadata_id].defined_execution_errors)
    return allowed_errors


def extract_metadata(
    context: grpc.ServicerContext, server: SilaServer, origin: Union[Property, Command]
) -> Dict[FullyQualifiedIdentifier, Any]:
    metadata = {}
    for entry in context.invocation_metadata():
        key: str = entry.key
        if re.fullmatch(
            f"sila/{FullyQualifiedIdentifierRegex.MetadataIdentifier}/bin", key.replace("-", "/"), flags=re.IGNORECASE
        ):
            key = key[5:-4].replace("-", "/")

            try:
                meta: Metadata = server.children_by_fully_qualified_identifier[FullyQualifiedIdentifier(key)]
            except KeyError:
                raise_as_rpc_error(UndefinedExecutionError(KeyError(f"Server has no metadata {key}")), context)

            try:
                value = meta.to_native_type(entry.value)
            except Exception as ex:
                raise_as_rpc_error(InvalidMetadata(f"Failed to deserialize metadata value for {key!r}: {ex}"), context)

            raw_affected_calls = getattr(
                server.feature_servicers[meta.parent_feature._identifier].implementation,
                f"get_calls_affected_by_{meta._identifier}",
            )()
            affected_calls = [
                FullyQualifiedIdentifier(call) if isinstance(call, str) else call.fully_qualified_identifier
                for call in raw_affected_calls
            ]
            if (
                origin.fully_qualified_identifier not in affected_calls
                and origin.parent_feature.fully_qualified_identifier not in affected_calls
            ):
                logger.warning(
                    f"Metadata ignored because call should not be affected: "
                    f"{str(meta.fully_qualified_identifier)} with value {value}"
                )
                continue

            metadata[meta.fully_qualified_identifier] = value

        else:
            logger.debug(f"Found non-SiLA metadata {key!r} with value {entry.value!r}")
    return metadata


def unpack_parameters(command: Command, request, context: ServicerContext) -> Tuple[Any, ...]:
    try:
        return command.parameters.to_native_type(request)
    except ValidationError as val_err:
        raise_as_rpc_error(val_err, context)


@contextmanager
def raises_rpc_errors(
    context: ServicerContext,
    allowed_defined_execution_errors: Optional[Iterable[DefinedExecutionErrorNode]] = None,
):
    try:
        yield
    except Exception as ex:
        if not isinstance(ex, SilaError):
            raise_as_rpc_error(UndefinedExecutionError(ex), context)

        allowed_error_identifiers = [err.fully_qualified_identifier for err in allowed_defined_execution_errors]
        if isinstance(ex, DefinedExecutionError) and ex.fully_qualified_identifier not in allowed_error_identifiers:
            raise_as_rpc_error(UndefinedExecutionError(ex), context)

        raise_as_rpc_error(ex, context)



import os
import sys
from os.path import join
from typing import Iterable, Optional

from sila2.code_generator.feature_generator import FeatureGenerator
from sila2.code_generator.template_environment import TemplateEnvironment
from sila2.framework import Feature


class PackageGenerator:
    def __init__(
        self,
        out_directory: str,
        package_name: Optional[str],
        features: Iterable[Feature],
        overwrite: bool,
    ):
        self.generate_package = package_name is not None

        if self.generate_package and not package_name.isidentifier():
            raise ValueError(f"Expected a valid package name, got {package_name!r}")

        self.out_directory = out_directory  # will be changed later
        self.package_name = package_name
        self.features = tuple(features)
        self.template_env = TemplateEnvironment("package")
        self.overwrite = overwrite

    def generate(self) -> None:
        if self.generate_package:
            # set out directory to package level
            os.makedirs(self.out_directory, exist_ok=True)
            original_out_directory = self.out_directory

            self._generate_file(
                join(self.out_directory, "setup.py"),
                self.template_env.get_template("setup").render(package_name=self.package_name),
            )
            self._generate_file(
                join(self.out_directory, "MANIFEST.in"),
                self.template_env.get_template("manifest").render(package_name=self.package_name),
            )

            # set out directory to package source code level
            self.out_directory = join(self.out_directory, self.package_name)
            os.makedirs(self.out_directory, exist_ok=True)

            self._generate_file(
                join(self.out_directory, "__main__.py"),
                self.template_env.get_template("main").render(package_name=self.package_name),
            )
            self._generate_file(
                join(self.out_directory, "__init__.py"),
                self.template_env.get_template("init").render(package_name=self.package_name),
            )
            self._generate_file(
                join(self.out_directory, "server.py"),
                self.template_env.get_template("server").render(package_name=self.package_name, features=self.features),
            )
            self.out_directory = join(self.out_directory, "generated")
        else:
            # set out directory to auto-generated source code level
            original_out_directory = self.out_directory

        os.makedirs(self.out_directory, exist_ok=True)
        self._generate_file(join(self.out_directory, "__init__.py"), "")
        self._generate_file(
            join(self.out_directory, "client.py"),
            self.template_env.get_template("client").render(package_name=self.package_name, features=self.features),
        )

        for feature in self.features:
            FeatureGenerator(feature).generate_all(join(self.out_directory, feature._identifier.lower()))

        os.system(f"{sys.executable} -m isort --line-length 120 --quiet --profile black {original_out_directory}")
        os.system(f"{sys.executable} -m black --line-length 120 --quiet {original_out_directory}")

    def _generate_file(self, out_filename: str, content: str) -> None:
        if not self.overwrite and os.path.isfile(out_filename):
            raise FileExistsError(f"File '{out_filename}' already exists. Set --overwrite to overwrite existing files.")
        with open(out_filename, "w") as fp:
            fp.write(content)



import os
import sys
from argparse import ArgumentParser
from typing import List, Optional

import sila2.code_generator
from sila2.code_generator.package_generator import PackageGenerator
from sila2.framework import Feature


def main(argv: Optional[List[str]] = None):
    parser = ArgumentParser(
        prog=sila2.code_generator.__name__,
        description="Generate Python package for a SiLA 2 server application from given feature definitions",
    )
    parser.add_argument("-p", "--package", help="Generate full package with given name", required=False)
    parser.add_argument("-o", "--out-dir", help="Output directory (default: '.')", default=".")
    parser.add_argument(
        "--overwrite",
        action="store_true",
        help="Override existing files (always true for feature-specific auto-generated files)",
    )
    parser.add_argument("--debug", action="store_true", help="Display more detailed error messages")
    parser.add_argument("feature_definitions", nargs="*", help="SiLA 2 feature definitions")

    args = parser.parse_args(argv)

    try:
        os.makedirs(args.out_dir, exist_ok=True)
        features = [
            Feature(open(feature_definition_file).read()) for feature_definition_file in args.feature_definitions
        ]
        PackageGenerator(args.out_dir, args.package, features, overwrite=args.overwrite).generate()
    except Exception as ex:
        if args.debug:
            raise ex

        print(f"{ex.__class__.__name__}: {ex}", file=sys.stderr)
        exit(1)
